import struct
import logging

# https://farsightsec.github.io/fstrm/

# Frame Streams Control Frame Format

# |------------------------------------|----------------------|
# | Data frame length                  | 4 bytes              |
# |------------------------------------|----------------------|
# | Control frame length               | 4 bytes              |
# |------------------------------------|----------------------|
# | Control frame type                 | 4 bytes              |
# |------------------------------------|----------------------|
# | Control frame content type         | 4 bytes (optional)   |
# |------------------------------------|----------------------|
# | Control frame content type length  | 4 bytes (optional)   |
# |------------------------------------|----------------------|
# | Payload                            | xx bytes             |     
# |------------------------------------|----------------------|

# Frame Streams Data Frame Format

# |------------------------------------|----------------------|
# | Data frame length                  | 4 bytes              |
# |------------------------------------|----------------------|
# | Payload - Protobuf                 | xx bytes             |
# |------------------------------------|----------------------|

FSTRM_DATA_FRAME = 0
FSTRM_CONTROL_ACCEPT = 1
FSTRM_CONTROL_START = 2
FSTRM_CONTROL_STOP = 3
FSTRM_CONTROL_READY = 4
FSTRM_CONTROL_FINISH = 5

FSTRM_CONTROL_FIELD_CONTENT_TYPE = 1
CONTENT_TYPE_DNSTAP = "protobuf:dnstap.Dnstap"

class FstrmHandler(object):
    """frame stream decoder/encoder"""
    def __init__(self):
        """init class"""
        self.buf = b''
        
        self.df_length = None
        self.cf_length = None
        
    def reset(self):
        """reset"""
        self.df_length = None
        self.cf_length = None
        
    def pending_nb_bytes(self):
        """returns number of bytes remaining pending"""
        if self.df_length is not None:
            if self.df_length > 0:
                return self.df_length - len(self.buf)

        if self.cf_length is not None:
            if self.cf_length > 0:
                return self.cf_length - len(self.buf)
                
        return 4
        
    def append(self, data):
        """append data to the buffer"""
        self.buf = b''.join([self.buf, data])
        
    def process(self):
        """process the buffer"""
        if self.df_length is None:
            # need more data ?
            if len(self.buf) < 4:
                return False 
        
            # enough data, decode frame length
            (self.df_length,) = struct.unpack("!I", self.buf[:4])
            self.buf = self.buf[4:]
   
        # control frame ?
        if self.df_length == 0:
            # need more data ?
            if len(self.buf) < 4:
                return False
            
            if self.cf_length is None:
                (self.cf_length,) = struct.unpack("!I", self.buf[:4])
                self.buf = self.buf[4:]
            
            # need more data ?
            if len(self.buf) < self.cf_length:
                return False
            # we have received enough data, the frame is complete
            return True
            
        else:
            # need more data ?
            if len(self.buf) < self.df_length:
                return False
            else:
                return True
                
    def decode(self):
        """decode frame"""
        is_cf = False
        
        # read frame from buffer
        if self.df_length == 0:
            frame = self.buf[:self.cf_length]
            self.buf = self.buf[self.cf_length:]
            is_cf = True
        else:
            frame = self.buf[:self.df_length]
            self.buf = self.buf[self.df_length:]

        # reset to process next frame
        self.df_length = None
        self.cf_length = None
        
        # data frame ?
        if not is_cf:
            return (FSTRM_DATA_FRAME, frame)
            
        # decode control frame
        (ctrl_type,) = struct.unpack("!I", frame[:4])
        frame = frame[4:]
        
        if ctrl_type == FSTRM_CONTROL_READY:
            (cf_ctype, cf_clength, ) = struct.unpack("!II", frame[:8])
            frame = frame[8:]
            
            if cf_ctype != FSTRM_CONTROL_FIELD_CONTENT_TYPE:
                raise Exception("control ready - content type invalid")
            if cf_clength > len(frame):
                raise Exception("control ready - content length invalid")
            
            cf_payload = frame[:cf_clength].decode()
            if cf_payload != CONTENT_TYPE_DNSTAP:
                raise Exception("control ready - dnstap protobuf expected")
                
        return (ctrl_type, frame)

    def encode(self, fs):
        """encode"""
        if fs == FSTRM_CONTROL_ACCEPT:
            frame = struct.pack('!IIIII', 0, 
                                len(CONTENT_TYPE_DNSTAP)+12,
                                FSTRM_CONTROL_ACCEPT,
                                FSTRM_CONTROL_FIELD_CONTENT_TYPE,
                                len(CONTENT_TYPE_DNSTAP))
            frame += CONTENT_TYPE_DNSTAP.encode()                   
        else:
            raise Exception("encode frame not yet supported: %s" % fs)
        
        return frame



import asyncio
import logging

clogger = logging.getLogger("dnstap_receiver.console")

def start_input(cfg, cb_onconnect, loop):
    clogger.debug("Input handler: unix socket")
    clogger.debug("Input handler: listening on %s" % cfg["path"])
  
    # asynchronous unix socket
    server = asyncio.start_unix_server(cb_onconnect, path=cfg["path"], loop=loop)
                                                  
    return server


import asyncio
import logging
import ssl

clogger = logging.getLogger("dnstap_receiver.console")

def start_input(cfg, cb_onconnect, loop):
    clogger.debug("Input handler: tcp socket")
    
    ssl_context = None
    if cfg["tls-support"]:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=cfg["tls-server-cert"], keyfile=["tls-server-key"])
        clogger.debug("Input handler - tls support enabled")
        
    clogger.debug("Input handler: listening on %s:%s" % (cfg["local-address"],cfg["local-port"])), 
    server = asyncio.start_server(cb_onconnect, cfg["local-address"],cfg["local-port"],
                                  ssl=ssl_context, loop=loop)
    return server


import json
import yaml

def convert_dnstap(fmt, tapmsg):
    """convert dnstap message"""
    if fmt == "text":
        msg_list = []
        msg_list.append("%s" % tapmsg["timestamp"])
        msg_list.append("%s" % tapmsg["identity"])
        msg_list.append("%s" % tapmsg["message"])
        msg_list.append("%s" % tapmsg["rcode"]) 
        msg_list.append("%s" % tapmsg["source-ip"])
        msg_list.append("%s" % tapmsg["source-port"])
        msg_list.append("%s" % tapmsg["family"])
        msg_list.append("%s" % tapmsg["protocol"])
        msg_list.append("%sb" % tapmsg["length"])
        msg_list.append("%s" % tapmsg["qname"])
        msg_list.append("%s" % tapmsg["rrtype"])
        msg_list.append("%s" % tapmsg["latency"])
        
        # geoip activated ?
        if "country" in tapmsg:
            msg_list.append("%s" % tapmsg["country"])
            msg_list.append("%s" % tapmsg["city"])
            
        msg = " ".join(msg_list)
        del msg_list
        return msg.encode()
        
    elif fmt == "json":
        msg = json.dumps(tapmsg)
        return msg.encode()
        
    elif fmt == "yaml":
        msg = yaml.dump(tapmsg)
        return msg.encode()
        
    else:
        return tapmsg
    


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dnstap.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='dnstap.proto',
  package='dnstap',
  syntax='proto2',
  serialized_pb=_b('\n\x0c\x64nstap.proto\x12\x06\x64nstap\"\x94\x01\n\x06\x44nstap\x12\x10\n\x08identity\x18\x01 \x01(\x0c\x12\x0f\n\x07version\x18\x02 \x01(\x0c\x12\r\n\x05\x65xtra\x18\x03 \x01(\x0c\x12!\n\x04type\x18\x0f \x02(\x0e\x32\x13.dnstap.Dnstap.Type\x12 \n\x07message\x18\x0e \x01(\x0b\x32\x0f.dnstap.Message\"\x13\n\x04Type\x12\x0b\n\x07MESSAGE\x10\x01\"\xac\x05\n\x07Message\x12\"\n\x04type\x18\x01 \x02(\x0e\x32\x14.dnstap.Message.Type\x12+\n\rsocket_family\x18\x02 \x01(\x0e\x32\x14.dnstap.SocketFamily\x12/\n\x0fsocket_protocol\x18\x03 \x01(\x0e\x32\x16.dnstap.SocketProtocol\x12\x15\n\rquery_address\x18\x04 \x01(\x0c\x12\x18\n\x10response_address\x18\x05 \x01(\x0c\x12\x12\n\nquery_port\x18\x06 \x01(\r\x12\x15\n\rresponse_port\x18\x07 \x01(\r\x12\x16\n\x0equery_time_sec\x18\x08 \x01(\x04\x12\x17\n\x0fquery_time_nsec\x18\t \x01(\x07\x12\x15\n\rquery_message\x18\n \x01(\x0c\x12\x12\n\nquery_zone\x18\x0b \x01(\x0c\x12\x19\n\x11response_time_sec\x18\x0c \x01(\x04\x12\x1a\n\x12response_time_nsec\x18\r \x01(\x07\x12\x18\n\x10response_message\x18\x0e \x01(\x0c\"\x95\x02\n\x04Type\x12\x0e\n\nAUTH_QUERY\x10\x01\x12\x11\n\rAUTH_RESPONSE\x10\x02\x12\x12\n\x0eRESOLVER_QUERY\x10\x03\x12\x15\n\x11RESOLVER_RESPONSE\x10\x04\x12\x10\n\x0c\x43LIENT_QUERY\x10\x05\x12\x13\n\x0f\x43LIENT_RESPONSE\x10\x06\x12\x13\n\x0f\x46ORWARDER_QUERY\x10\x07\x12\x16\n\x12\x46ORWARDER_RESPONSE\x10\x08\x12\x0e\n\nSTUB_QUERY\x10\t\x12\x11\n\rSTUB_RESPONSE\x10\n\x12\x0e\n\nTOOL_QUERY\x10\x0b\x12\x11\n\rTOOL_RESPONSE\x10\x0c\x12\x10\n\x0cUPDATE_QUERY\x10\r\x12\x13\n\x0fUPDATE_RESPONSE\x10\x0e*#\n\x0cSocketFamily\x12\x08\n\x04INET\x10\x01\x12\t\n\x05INET6\x10\x02*\"\n\x0eSocketProtocol\x12\x07\n\x03UDP\x10\x01\x12\x07\n\x03TCP\x10\x02')
)

_SOCKETFAMILY = _descriptor.EnumDescriptor(
  name='SocketFamily',
  full_name='dnstap.SocketFamily',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INET', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INET6', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=862,
  serialized_end=897,
)
_sym_db.RegisterEnumDescriptor(_SOCKETFAMILY)

SocketFamily = enum_type_wrapper.EnumTypeWrapper(_SOCKETFAMILY)
_SOCKETPROTOCOL = _descriptor.EnumDescriptor(
  name='SocketProtocol',
  full_name='dnstap.SocketProtocol',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UDP', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TCP', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=899,
  serialized_end=933,
)
_sym_db.RegisterEnumDescriptor(_SOCKETPROTOCOL)

SocketProtocol = enum_type_wrapper.EnumTypeWrapper(_SOCKETPROTOCOL)
INET = 1
INET6 = 2
UDP = 1
TCP = 2


_DNSTAP_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='dnstap.Dnstap.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MESSAGE', index=0, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=154,
  serialized_end=173,
)
_sym_db.RegisterEnumDescriptor(_DNSTAP_TYPE)

_MESSAGE_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='dnstap.Message.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='AUTH_QUERY', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AUTH_RESPONSE', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESOLVER_QUERY', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESOLVER_RESPONSE', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_QUERY', index=4, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_RESPONSE', index=5, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORWARDER_QUERY', index=6, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORWARDER_RESPONSE', index=7, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STUB_QUERY', index=8, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STUB_RESPONSE', index=9, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOOL_QUERY', index=10, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOOL_RESPONSE', index=11, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UPDATE_QUERY', index=12, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UPDATE_RESPONSE', index=13, number=14,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=583,
  serialized_end=860,
)
_sym_db.RegisterEnumDescriptor(_MESSAGE_TYPE)


_DNSTAP = _descriptor.Descriptor(
  name='Dnstap',
  full_name='dnstap.Dnstap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='identity', full_name='dnstap.Dnstap.identity', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='dnstap.Dnstap.version', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra', full_name='dnstap.Dnstap.extra', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='dnstap.Dnstap.type', index=3,
      number=15, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='dnstap.Dnstap.message', index=4,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DNSTAP_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=173,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='dnstap.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='dnstap.Message.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='socket_family', full_name='dnstap.Message.socket_family', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='socket_protocol', full_name='dnstap.Message.socket_protocol', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_address', full_name='dnstap.Message.query_address', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response_address', full_name='dnstap.Message.response_address', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_port', full_name='dnstap.Message.query_port', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response_port', full_name='dnstap.Message.response_port', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_time_sec', full_name='dnstap.Message.query_time_sec', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_time_nsec', full_name='dnstap.Message.query_time_nsec', index=8,
      number=9, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_message', full_name='dnstap.Message.query_message', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query_zone', full_name='dnstap.Message.query_zone', index=10,
      number=11, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response_time_sec', full_name='dnstap.Message.response_time_sec', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response_time_nsec', full_name='dnstap.Message.response_time_nsec', index=12,
      number=13, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='response_message', full_name='dnstap.Message.response_message', index=13,
      number=14, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MESSAGE_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=176,
  serialized_end=860,
)

_DNSTAP.fields_by_name['type'].enum_type = _DNSTAP_TYPE
_DNSTAP.fields_by_name['message'].message_type = _MESSAGE
_DNSTAP_TYPE.containing_type = _DNSTAP
_MESSAGE.fields_by_name['type'].enum_type = _MESSAGE_TYPE
_MESSAGE.fields_by_name['socket_family'].enum_type = _SOCKETFAMILY
_MESSAGE.fields_by_name['socket_protocol'].enum_type = _SOCKETPROTOCOL
_MESSAGE_TYPE.containing_type = _MESSAGE
DESCRIPTOR.message_types_by_name['Dnstap'] = _DNSTAP
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.enum_types_by_name['SocketFamily'] = _SOCKETFAMILY
DESCRIPTOR.enum_types_by_name['SocketProtocol'] = _SOCKETPROTOCOL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dnstap = _reflection.GeneratedProtocolMessageType('Dnstap', (_message.Message,), dict(
  DESCRIPTOR = _DNSTAP,
  __module__ = 'dnstap_pb2'
  # @@protoc_insertion_point(class_scope:dnstap.Dnstap)
  ))
_sym_db.RegisterMessage(Dnstap)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGE,
  __module__ = 'dnstap_pb2'
  # @@protoc_insertion_point(class_scope:dnstap.Message)
  ))
_sym_db.RegisterMessage(Message)


# @@protoc_insertion_point(module_scope)




import dns.exception
import dns.opcode
import dns.flags

# waiting fix with dnspython 2.1
# will be removed in the future
class _WireReader(dns.message._WireReader):
    def read(self):
        """issue fixed - waiting fix with dnspython 2.1"""
        if self.parser.remaining() < 12:
            raise dns.message.ShortHeader
        (id, flags, qcount, ancount, aucount, adcount) = \
            self.parser.get_struct('!HHHHHH')
        factory = dns.message._message_factory_from_opcode(dns.opcode.from_flags(flags))
        self.message = factory(id=id)
        self.message.flags = flags
        self.initialize_message(self.message)
        self.one_rr_per_rrset = \
            self.message._get_one_rr_per_rrset(self.one_rr_per_rrset)
        self._get_question(dns.message.MessageSection.QUESTION, qcount)
        
        return self.message

# waiting fix with dnspython 2.1
# will be removed in the future
def from_wire(wire, question_only=True):
    """decode wire message - waiting fix with dnspython 2.1"""
    raise_on_truncation=False
    def initialize_message(message):
        message.request_mac = b''
        message.xfr = False
        message.origin = None
        message.tsig_ctx = None

    reader = _WireReader(wire, initialize_message, question_only=question_only,
                 one_rr_per_rrset=False, ignore_trailing=False,
                 keyring=None, multi=False)
    try:
        m = reader.read()
    except dns.exception.FormError:
        if reader.message and (reader.message.flags & dns.flags.TC) and \
           raise_on_truncation:
            raise dns.message.Truncated(message=reader.message)
        else:
            raise
    # Reading a truncated message might not have any errors, so we
    # have to do this check here too.
    if m.flags & dns.flags.TC and raise_on_truncation:
        raise dns.message.Truncated(message=m)

    return m
    


import argparse
import logging
import asyncio
import socket
import yaml
import sys
import re
import ssl
import pkgutil
import ipaddress

from datetime import datetime, timezone

import geoip2.database

# python3 -m pip dnspython
import dns.rcode
import dns.rdatatype
import dns.message

# create default logger for the dnstap receiver
clogger = logging.getLogger("dnstap_receiver.console")

# import framestreams and dnstap protobuf decoder
from dnstap_receiver.codecs import dnstap_pb2 
from dnstap_receiver.codecs import fstrm 

# import all inputs
from dnstap_receiver.inputs import input_unixsocket
from dnstap_receiver.inputs import input_tcpsocket

# import all outputs
from dnstap_receiver.outputs import output_stdout
from dnstap_receiver.outputs import output_file
from dnstap_receiver.outputs import output_syslog
from dnstap_receiver.outputs import output_tcp
from dnstap_receiver.outputs import output_metrics

from dnstap_receiver import api_server
from dnstap_receiver import statistics
from dnstap_receiver import dnspython_patch

class UnknownValue:
    name = "-"

DNSTAP_TYPE = dnstap_pb2._MESSAGE_TYPE.values_by_number
DNSTAP_FAMILY = dnstap_pb2._SOCKETFAMILY.values_by_number
DNSTAP_PROTO = dnstap_pb2._SOCKETPROTOCOL.values_by_number  

DFLT_LISTEN_IP = "0.0.0.0"
DFLT_LISTEN_PORT = 6000

# command line arguments definition
parser = argparse.ArgumentParser()
parser.add_argument("-l", 
                    help="IP of the dnsptap server to receive dnstap payloads (default: %(default)r)",
                    default=DFLT_LISTEN_IP)
parser.add_argument("-p", type=int,
                    help="Port the dnstap receiver is listening on (default: %(default)r)",
                    default=DFLT_LISTEN_PORT)               
parser.add_argument("-u", help="read dnstap payloads from unix socket")
parser.add_argument('-v', action='store_true', help="verbose mode")   
parser.add_argument("-c", help="external config file")   

async def cb_ondnstap(dnstap_decoder, payload, cfg, queues_list, stats, geoip_reader):
    """on dnstap"""
    # decode binary payload
    dnstap_decoder.ParseFromString(payload)
    dm = dnstap_decoder.message
    
    if cfg["trace"]["dnstap"]:
        dns_pkt = dm.query_message if (dm.type % 2 ) == 1 else dm.response_message
        clogger.debug("%s\n%s\n\n" % (dm,dns.message.from_wire(dns_pkt)) )

    # filtering by dnstap identity ?
    tap_ident = dnstap_decoder.identity.decode()
    if not len(tap_ident):
        tap_ident = UnknownValue.name
    if cfg["filter"]["dnstap-identities"] is not None:
        if re.match(cfg["filter"]["dnstap-identities"], dnstap_decoder.identity.decode()) is None:
            del dm
            return
            
    tap = { "identity": tap_ident, "qname": UnknownValue.name, "rrtype": UnknownValue.name, 
            "query-type": UnknownValue.name, "source-ip": UnknownValue.name}
    
    # decode type message
    tap["message"] = DNSTAP_TYPE.get(dm.type, UnknownValue).name
    tap["family"] = DNSTAP_FAMILY.get(dm.socket_family, UnknownValue).name
    tap["protocol"] = DNSTAP_PROTO.get(dm.socket_protocol, UnknownValue).name

    # decode query address
    qaddr = dm.query_address
    if len(qaddr) and dm.socket_family == 1:
        # condition for coredns, address is 16 bytes long so keept only 4 bytes
        qaddr = qaddr[12:] if len(qaddr) == 16 else qaddr
        # convert ip to string
        tap["source-ip"] = socket.inet_ntoa(qaddr)
    if len(qaddr) and dm.socket_family == 2:
        tap["source-ip"] = socket.inet_ntop(socket.AF_INET6, qaddr)
    tap["source-port"] = dm.query_port
    if tap["source-port"] == 0:
        tap["source-port"] = UnknownValue.name
        
    # handle query message
    # todo catching dns.message.ShortHeader exception
    # can occured with coredns if the full argument is missing
    if (dm.type % 2 ) == 1 :
        dnstap_parsed = dnspython_patch.from_wire(dm.query_message,
                                  question_only=True)                 
        tap["length"] = len(dm.query_message)
        d1 = dm.query_time_sec +  (round(dm.query_time_nsec ) / 1000000000)
        tap["timestamp"] = datetime.fromtimestamp(d1, tz=timezone.utc).isoformat()
        tap["type"] = "query"
        latency = 0.0
        
    # handle response message
    if (dm.type % 2 ) == 0 :
        dnstap_parsed = dnspython_patch.from_wire(dm.response_message,
                                  question_only=True)
        tap["length"] = len(dm.response_message)
        d2 = dm.response_time_sec + (round(dm.response_time_nsec ) / 1000000000) 
        tap["timestamp"] = datetime.fromtimestamp(d2, tz=timezone.utc).isoformat()
        tap["type"] = "response"

        # compute latency 
        d1 = dm.query_time_sec +  (round(dm.query_time_nsec ) / 1000000000)
        latency = round(d2-d1,3)
    
    tap["latency"] = latency
        
    # common params
    if len(dnstap_parsed.question):
        tap["qname"] = dnstap_parsed.question[0].name.to_text()
        tap["rrtype"] = dns.rdatatype.to_text(dnstap_parsed.question[0].rdtype)
    tap["rcode"] = dns.rcode.to_text(dnstap_parsed.rcode())
    tap["id"] = dnstap_parsed.id
    tap["flags"] = dns.flags.to_text(dnstap_parsed.flags)

    # filtering by qname ?
    if cfg["filter"]["qname-regex"] is not None:
        if re.match(cfg["filter"]["qname-regex"], tap["qname"]) is None:
            del dm; del tap;
            return

    # geoip support 
    if geoip_reader is not None:
        try:
            response = geoip_reader.city(tap["source-ip"])
            if cfg["geoip"]["country-iso"]:
                tap["country"] = response.country.iso_code
            else:
                tap["country"] = response.country.name
            if response.city.name is not None:
                tap["city"] = response.city.name
            else:
                tap["city"] = UnknownValue.name
        except Exception as e:
            tap["country"] = UnknownValue.name
            tap["city"] = UnknownValue.name
            
    # update metrics 
    stats.record(tap=tap)
        
    # append the dnstap message to the queue
    for q in queues_list:
        q.put_nowait(tap)
    
async def cb_onconnect(reader, writer, cfg, queues_list, stats, geoip_reader):
    """callback when a connection is established"""
    # get peer name
    peername = writer.get_extra_info('peername')
    if not len(peername):
        peername = "(unix-socket)"
    clogger.debug(f"Input handler: new connection from {peername}")

    # access control list check
    if len(writer.get_extra_info('peername')):
        acls_network = []
        for a in cfg["input"]["tcp-socket"]["access-control-list"]:
            acls_network.append(ipaddress.ip_network(a))
            
        acl_allow = False
        for acl in acls_network:
            if ipaddress.ip_address(peername[0]) in acl:
                acl_allow = True
        
        if not acl_allow:
            writer.close()
            clogger.debug("Input handler: checking acl refused")
            return
        
        clogger.debug("Input handler: checking acl allowed")
        
    # prepare frame streams decoder
    fstrm_handler = fstrm.FstrmHandler()
    loop = asyncio.get_event_loop()
    dnstap_decoder = dnstap_pb2.Dnstap()

    try: 
        # syntax only works with python 3.8
        # while data := await reader.read(fstrm_handler.pending_nb_bytes()) 
        running = True
        while running:
            # read bytes
            data = await reader.read(fstrm_handler.pending_nb_bytes()) 
            if not len(data):
                running = False
                break
                
            # append data to the buffer
            fstrm_handler.append(data=data)
            
            # process the buffer, check if we have received a complete frame ?
            if fstrm_handler.process():
                # Ok, the frame is complete so let's decode it
                fs, payload  = fstrm_handler.decode()

                # handle the DATA frame
                if fs == fstrm.FSTRM_DATA_FRAME:
                    loop.create_task(cb_ondnstap(dnstap_decoder, payload, cfg, queues_list, stats, geoip_reader))
                    
                # handle the control frame READY
                if fs == fstrm.FSTRM_CONTROL_READY:
                    clogger.debug(f"Input handler: control ready received from {peername}")
                    ctrl_accept = fstrm_handler.encode(fs=fstrm.FSTRM_CONTROL_ACCEPT)
                    # respond with accept only if the content type is dnstap
                    writer.write(ctrl_accept)
                    await writer.drain()
                    clogger.debug(f"Input handler: sending control accept to {peername}")
                    
                # handle the control frame READY
                if fs == fstrm.FSTRM_CONTROL_START:
                    clogger.debug(f"Input handler: control start received from {peername}")
   
                # handle the control frame STOP
                if fs == fstrm.FSTRM_CONTROL_STOP:
                    clogger.debug(f"Input handler: control stop received from {peername}")
                    fstrm_handler.reset()           
    except asyncio.CancelledError:
        clogger.debug(f'Input handler: {peername} - closing connection.')
        writer.close()
        await writer.wait_closed()
    except asyncio.IncompleteReadError:
        clogger.debug(f'Input handler: {peername} - disconnected')
    finally:
        clogger.debug(f'Input handler: {peername} - closed')

def merge_cfg(u, o):
    """merge config"""
    for k,v in u.items():
        if k in o:
            if isinstance(v, dict):
                merge_cfg(u=v,o=o[k])
            else:
                o[k] = v
   
def setup_config(args):
    """load default config and update it with arguments if provided"""
    # set default config
    try:
        cfg =  yaml.safe_load(pkgutil.get_data(__package__, 'dnstap.conf')) 
    except FileNotFoundError:
        print("default config file not found")
        sys.exit(1)
    except yaml.parser.ParserError:
        print("invalid default yaml config file")
        sys.exit(1)

    # overwrite config with external file ?    
    if args.c:
        try:
            with open(args.c) as file:
                merge_cfg(u=yaml.safe_load(file),o=cfg)
        except FileNotFoundError:
            print("external config file not found")
            sys.exit(1)
        except yaml.parser.ParserError:
            print("external invalid yaml config file")
            sys.exit(1)

    # update default config with command line arguments
    if args.v:
        cfg["trace"]["verbose"] = args.v    
    if args.u is not None:
        cfg["input"]["unix-socket"]["path"] = args.u
    if args.l != DFLT_LISTEN_IP:
        cfg["input"]["tcp-socket"]["local-address"] = args.l
    if args.l != DFLT_LISTEN_PORT:
        cfg["input"]["tcp-socket"]["local-port"] = args.p

    return cfg
    
def setup_logger(cfg):
    """setup main logger"""

    loglevel = logging.DEBUG if cfg["verbose"] else logging.INFO
    logfmt = '%(asctime)s %(levelname)s %(message)s'
    
    clogger.setLevel(loglevel)
    clogger.propagate = False
    
    if cfg["file"] is None:
        lh = logging.StreamHandler(stream=sys.stdout )
    else:
        lh = logging.FileHandler(cfg["file"])
    lh.setLevel(loglevel)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    clogger.addHandler(lh)
    
def setup_outputs(cfg, stats, loop):
    """setup outputs"""
    conf = cfg["output"]

    queues_list = []
    if conf["syslog"]["enable"]:
        if not output_syslog.checking_conf(cfg=conf["syslog"]): return
        queue_syslog = asyncio.Queue()
        queues_list.append(queue_syslog)
        loop.create_task(output_syslog.handle(conf["syslog"], queue_syslog, stats))    

    if conf["tcp-socket"]["enable"]:
        if not output_tcp.checking_conf(cfg=conf["tcp-socket"]): return
        queue_tcpsocket = asyncio.Queue()
        queues_list.append(queue_tcpsocket)
        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue_tcpsocket, stats))
                                               
    if conf["file"]["enable"]:
        if not output_file.checking_conf(cfg=conf["file"]): return
        queue_file = asyncio.Queue()
        queues_list.append(queue_file)
        loop.create_task(output_file.handle(conf["file"], queue_file, stats))
                                              
    if conf["stdout"]["enable"]:
        if not output_stdout.checking_conf(cfg=conf["stdout"]): return
        queue_stdout = asyncio.Queue()
        queues_list.append(queue_stdout)
        loop.create_task(output_stdout.handle(conf["stdout"], queue_stdout, stats))

    if conf["metrics"]["enable"]:
        if not output_metrics.checking_conf(cfg=conf["metrics"]): return
        queue_metrics = asyncio.Queue()
        queues_list.append(queue_metrics)
        loop.create_task(output_metrics.handle(conf["metrics"], queue_metrics, stats))

    return queues_list
    
def setup_inputs(args, cfg, queues_list, stats, loop, geoip_reader):
    """setup inputs"""
    # define callback on new connection
    cb_lambda = lambda r, w: cb_onconnect(r, w, cfg, queues_list, stats, geoip_reader)
    
    # asynchronous unix socket
    if cfg["input"]["unix-socket"]["path"] is not None:
        socket_server = input_unixsocket.start_input(cfg=cfg["input"]["unix-socket"], cb_onconnect=cb_lambda, loop=loop)

    # default mode: asynchronous tcp socket
    else:
        socket_server = input_tcpsocket.start_input(cfg=cfg["input"]["tcp-socket"], cb_onconnect=cb_lambda, loop=loop)
                                             
    # run until complete
    loop.run_until_complete(socket_server)
    
def setup_api(cfg, stats, loop):
    """setup web api"""
    if cfg["web-api"]["enable"]:
        api_svr = api_server.create_server(loop, cfg=cfg["web-api"], 
                                           stats=stats, cfg_stats=cfg["statistics"])
        loop.run_until_complete(api_svr)

def setup_geoip(cfg):
    if not cfg["enable"]: return None
    if cfg["city-database"] is None: return None
    
    reader = None
    try:
        reader = geoip2.database.Reader(cfg["city-database"])
    except Exception as e:
        clogger.error("geoip setup: %s" % e)
        
    return reader
    
def start_receiver():
    """start dnstap receiver"""
    # Handle command-line arguments.
    args = parser.parse_args()

    # setup config
    cfg = setup_config(args=args)
  
    # setup logging
    setup_logger(cfg=cfg["trace"])

    # setup geoip if enabled 
    geoip_reader = setup_geoip(cfg=cfg["geoip"])
    
    # add debug message if external config is used
    if args.c: clogger.debug("External config file loaded")
    
    # start receiver and get event loop
    clogger.debug("Start receiver...")
    loop = asyncio.get_event_loop()
    stats = statistics.Statistics(cfg=cfg["statistics"])
    loop.create_task(statistics.watcher(stats))
    
    # prepare outputs
    queues_list = setup_outputs(cfg, stats, loop)
    
    # prepare inputs
    setup_inputs(args, cfg, queues_list, stats, loop, geoip_reader)

    # start the http api
    setup_api(cfg, stats, loop)

    # run event loop 
    try:
       loop.run_forever()
    except KeyboardInterrupt:
        pass

    # close geoip
    if geoip_reader is not None: geoip_reader.close()



from collections import Counter
from collections import defaultdict

import asyncio
import re
from tlds import tld_set

# watcher for compute qps
async def watcher(statistics):
    """watcher for statistics"""
    while True:
        # sleep during one second
        await asyncio.sleep(1)
        
        # refresh counters and compute qps every interval
        statistics.update_counters()
        statistics.compute_qps()
    
class StatsStream:
    def __init__(self, name, stats):
        """constructor"""
        self.stats = stats
        self.name = name

        self.bufq = defaultdict(Counter)
        self.bufr = defaultdict(Counter)
        self.bufi = defaultdict(Counter)

        self.prev_qr = 0
        
        self.cnts = Counter()
        self.cnts_rcode = Counter()
        self.cnts_rrtype = Counter()

        self.cnts_tlds = Counter()
        
        self.cnts_latency = Counter()
        
    def record(self, tap):
        """record only response dnstap message"""
        qname = tap["qname"]; srcip = tap["source-ip"]; 
        qr = tap["type"]; rcode = tap["rcode"]; rrtype = tap["rrtype"]
        
        # transform qname to lowercase ?
        if self.stats.cfg["qname-lowercase"]:
            qname = qname.lower()

        # count number of hit and bytes for each source ip
        self.bufi[srcip]["hit"] += 1
        self.bufi[srcip]["length"] += tap["length"]
        self.cnts["clients"] = len(self.bufi)
        
        # count number of dnstap query or response.
        self.cnts[qr] += 1
        
        # count number of dnstap according to the protocol and family
        self.cnts["%s/%s" % (qr,tap["protocol"].lower())] += 1
        self.cnts["%s/%s" % (qr,tap["family"].lower())] += 1

        # count total of bytes
        self.cnts["%s/bytes" % qr] += tap["length"]
        
        # prepare the buffer according to the dnstap message
        buf = self.bufq if qr == "query" else self.bufr 
        
        # count number of hit and byte for each qname
        buf[qname]["hit"] += 1
        buf[qname]["length"] += tap["length"]
        
        # count number of rcode and rrtype for each qname
        buf[qname][rcode.lower()] += 1
        buf[qname][rrtype.lower()] += 1
        
        # count number of rcode and rrtype for each qname
        self.cnts_rcode["%s/%s" % (qr,rcode.lower())] += 1
        self.cnts_rrtype["%s/%s" % (qr,rrtype.lower())] += 1

        # count number of unique domains
        self.cnts["domains"] = len(buf)

        # count tld
        tld_matched = list( filter(lambda x: x in tld_set, qname.rsplit(".", 2)) )
        if len(tld_matched):
            self.cnts_tlds["%s/%s" % (qr,tld_matched[-1])] += 1
        
        # latency
        if tap["latency"] <= 0.001:
            self.cnts_latency["%s/latency0_1" % qr ] += 1
        if 0.001 < tap["latency"] <= 0.010 :
            self.cnts_latency["%s/latency1_10" % qr ] += 1
        if 0.010 < tap["latency"] <= 0.050 :
            self.cnts_latency["%s/latency10_50" % qr ] += 1
        if 0.050 < tap["latency"] <= 0.100 :
            self.cnts_latency["%s/latency50_100" % qr ] += 1
        if 0.100 < tap["latency"] <= 1.000 :
            self.cnts_latency["%s/latency100_1000" % qr ] += 1    
        if tap["latency"] > 1.000 :
            self.cnts_latency["%s/latency_slow" % qr ] += 1    
            
    def reset(self):
        """reset the stream"""
        # reset all counters and buffers
        self.bufi.clear()
        self.bufq.clear()
        self.bufr.clear()
        
        self.cnts.clear()
        self.cnts_rcode.clear()
        self.cnts_rrtype.clear()
        
        self.cnts_tlds.clear()
        self.cnts_latency.clear()
        
        self.prev_qr = 0

    def compute_qps(self):
        """compute qps query/qps and response/qps"""
        cur_qr = self.cnts.get("query", 0)
        if cur_qr == 0: return

        qps = cur_qr - self.prev_qr
        if qps < 0: qps = 0
        self.cnts["qps"]  = qps
        self.prev_qr = cur_qr
        
class Statistics:
    def __init__(self, cfg):
        """constructor"""
        self.cfg = cfg
        self.streams = {}
        
        # Counter({'query/response': <int>, 'query|response/udp|tcp': <int>, 
        # 'query|response/inet|inet6': <int>, 'domains': <int>, 'clients': <int>})
        self.cnts = Counter()
        # Counter({'query|response/<rcode>': <int>})
        self.cnts_rcode = Counter()
        # Counter({'query|response/<rrtype>': <int>})
        self.cnts_rrtype = Counter()
        
        # Counter({'query|response/<tlds>': <int>})
        self.cnts_tlds = Counter()
        # Counter({'response/latency0_10': <int>})
        self.cnts_latency = Counter()
        
        self.global_qps = Counter()
        
    def record(self, tap):
        """record dnstap message"""
        if tap["identity"] not in self.streams:
            s = StatsStream(name=tap["identity"], stats=self)
            self.streams[tap["identity"]] = s
        self.streams[tap["identity"]].record(tap=tap)

    def update_counters(self):
        """create global counters"""
        # update global counters
        self.cnts.clear()
        self.cnts_rcode.clear()
        self.cnts_rrtype.clear()
        self.cnts_tlds.clear()
        self.cnts_latency.clear()
        
        qnames = set()
        ips = set()
        for s in self.streams:
            ips.update(set(self.streams[s].bufi))
            
            qnames.update(set(self.streams[s].bufr))
            qnames.update(set(self.streams[s].bufq))
            
            self.cnts.update(self.streams[s].cnts)
            self.cnts_rcode.update(self.streams[s].cnts_rcode)
            self.cnts_rrtype.update(self.streams[s].cnts_rrtype)
  
            self.cnts_tlds.update(self.streams[s].cnts_tlds)
            self.cnts_latency.update(self.streams[s].cnts_latency)
            
        self.cnts["clients"] = len(ips)
        self.cnts["domains"] = len(qnames)
  
    def reset(self):
        """reset all streams"""
        # reset all counters and buffer in all streams
        for s in self.get_streams():
            s.reset()

        # reset global counters
        self.cnts.clear()
        self.cnts_rcode.clear()
        self.cnts_rrtype.clear()
        self.global_qps.clear()

        self.cnts_tlds.clear()
        self.cnts_latency.clear()
        
    def get_streams(self, stream=None):
        """return list of stream object"""
        if stream is None:
            return list(self.streams.values())

        s = self.streams.get(stream)
        if s is None:
            return []
        else:
            return [s]
         
    def get_nameslist(self):
        """return stream name in a list"""
        return list(self.streams.keys())

    def compute_qps(self):
        """create some global counters"""
        self.global_qps.clear()
        
        for s in self.get_streams():
            s.compute_qps()
            self.global_qps.update({"qps": s.cnts["qps"]})
        
        self.cnts["qps"] = self.global_qps["qps"]
      
    def get_counters(self, stream=None, filters=[]):
        """return all counters"""
        # get computed counters according to the stream
        # if the stream is not found, return the global counters
        s = self.streams.get(stream)
        _cnt = Counter()
        if s is None:
            _cnt.update(self.cnts)
            _cnt.update(self.cnts_rcode)
            _cnt.update(self.cnts_rrtype)
            _cnt.update(self.cnts_tlds)
            _cnt.update(self.cnts_latency)
        else:
            _cnt.update(s.cnts)
            _cnt.update(s.cnts_rcode)
            _cnt.update(s.cnts_rrtype)
            _cnt.update(s.cnts_tlds)
            _cnt.update(s.cnts_latency)
            
        # set counters
        c = {}
        for f in filters:
            c[f] = _cnt.get(f,0)
            
        return c

    def top_rrtypes(self, n, stream=None):
        """return top- hit/response|query"""
        top = {}
        s = self.streams.get(stream)
        cnt = s.cnts_rrtype if s is not None else self.cnts_rrtype
            
        for qr in [ "query", "response" ]:
            cnt_ = Counter(dict(filter(lambda x:x[0].startswith("%s" % qr), cnt.items())))
            top["hit/%s" % qr] = cnt_.most_common(n)
        return top
        
    def top_rcodes(self, n, stream=None):
        """return top- hit/response|query"""
        top = {}
        s = self.streams.get(stream)
        cnt = s.cnts_rcode if s is not None else self.cnts_rcode 
        for qr in [ "query", "response" ]:
            cnt_ = Counter(dict(filter(lambda x:x[0].startswith("%s" % qr), cnt.items())))
            top["hit/%s" % qr] = cnt_.most_common(n)
        return top
        
    def top_tlds(self, n, stream):
        """return top tld"""
        top = {}
        s = self.streams.get(stream)
        cnt = s.cnts_tlds if s is not None else self.cnts_tlds   
        for qr in [ "query", "response" ]:
            cnt_ = Counter(dict(filter(lambda x:x[0].startswith("%s" % qr), cnt.items())))
            top["hit/%s" % qr] = cnt_.most_common(n)
        return top
        
    def top_clients(self, n, stream):
        """return top clients"""
        top = {}
        for flag in [ "hit", "length" ]:
            cnt_ = Counter()
            for s in self.get_streams(stream=stream):
                for ip in s.bufi: 
                    cnt_.update({ip:s.bufi[ip][flag]})         
            top["%s/ip" % flag] = cnt_.most_common(n)
        return top
    
    def top_domains(self, n, stream, filters=[]):
        """return top domains"""
        top = {}
        for flag in filters:
            
            if "/" not in flag: break
            by, qr = flag.split("/")
            if qr not in ["query", "response"]: break
            
            top_list = []
            for s in self.get_streams(stream=stream):
                s_ = s.bufq if qr == "query" else s.bufr
                f = filter(lambda x:x[1].get(by, 0) > 0, s_.items())
                top_list.extend( sorted(f, key= lambda x:x[1][by], reverse=True)[:n] )

            # remove duplicate
            cnt_ = Counter()
            for (domain, counters) in top_list:
                cnt_.update({domain: counters[by]})
            top[flag] = cnt_.most_common(n)
        return top


#!/usr/bin/python

import setuptools

with open("./dnstap_receiver/__init__.py", "r") as fh:
    for line in fh.read().splitlines():
        if line.startswith('__version__'):
            PKG_VERSION = line.split('"')[1]
            
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
    
KEYWORDS = ('dnstap receiver client json yaml text')

setuptools.setup(
    name="dnstap_receiver",
    version=PKG_VERSION,
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="Python Dnstap receiver",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/dnstap_receiver",
    packages=['dnstap_receiver'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points={'console_scripts': ['dnstap_receiver = dnstap_receiver.receiver:start_receiver']},
    install_requires=[
        "dnspython",
        "protobuf",
        "pyyaml",
        "aiohttp",
        "geoip2",
        "tlds"
    ]
)



import logging
import logging.handlers
import sys

clogger = logging.getLogger("dnstap_receiver.console")
file_logger = logging.getLogger("dnstap_receiver.output.file")

from dnstap_receiver.outputs import transform

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: file")
    
    return True
    
def setup_logger(cfg):
    """setup loggers"""
    logfmt = '%(message)s'
    max_bytes = int(cfg["file-max-size"].split('M')[0]) * 1024 * 1024
    
    file_logger.setLevel(logging.INFO)
    file_logger.propagate = False
    
    lh = logging.handlers.RotatingFileHandler(
        cfg["file"],
        maxBytes=max_bytes,
        backupCount=cfg["file-count"]
    )
    lh.setLevel(logging.INFO)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    file_logger.addHandler(lh)
    
async def handle(output_cfg, queue, metrics):
    """stdout output handler"""

    # init output logger
    setup_logger(output_cfg)
    
    while True:
        # read item from queue
        tapmsg = await queue.get()
        
        # convert dnstap message
        msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)
        
        # print to stdout
        file_logger.info(msg.decode())
        
        # all done
        queue.task_done()